<?php

namespace Drupal\unique_field_ajax;

use Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException;
use Drupal\Component\Plugin\Exception\PluginNotFoundException;
use Drupal\Core\Entity\ContentEntityType;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Field\BaseFieldDefinition;
use Drupal\Core\Field\Entity\BaseFieldOverride;
use Drupal\Core\Form\FormStateInterface;
use Drupal\node\Entity\NodeType;
use Drupal\field\Entity\FieldConfig;
use Drupal\Core\StringTranslation\StringTranslationTrait;

/**
 * Handles all the form elements for Unique Field Ajax.
 */
class UniqueFieldAjaxFormFieldsService {
  use StringTranslationTrait;

  // Third party settings key.
  const UNIQUE_THIRD_PARTY_SETTINGS_KEY = 'unique_field_ajax';

  // Extra third party settings (TPS).
  const TPS_FIELD_NAME = 'field_name';
  const TPS_FIELD_LABEL = 'field_label';
  const TPS_MAIN_PROPERTY_NAME = 'main_property_name';

  // Field keys MUST start with 'FIELD_'.
  const FIELD_UNIQUE = 'unique';
  const FIELD_PER_LANG = 'per_lang';
  const FIELD_CASE_SENSITIVE = 'case_sensitive';
  const FIELD_USE_AJAX = 'use_ajax';
  const FIELD_NO_ENFORCE = 'no_enforce';
  const FIELD_MESSAGE_WARNING = 'message_warning';
  const FIELD_MESSAGE = 'message';

  // Tokens for replacement.
  const TOKEN_LABEL = '%label';
  const TOKEN_LINK = '%link';
  const TOKEN_VALUE = '%value';

  // Default warning/error messages.
  const DEFAULT_WARNING_MESSAGE = "Field '" . self::TOKEN_LABEL . "' <em>should</em> be unique.\r\nThe value '" . self::TOKEN_VALUE . "' can also be found on " . self::TOKEN_LINK . ".";
  const DEFAULT_ERROR_MESSAGE = "Field '" . self::TOKEN_LABEL . "' <strong>must</strong> be unique.\r\nThe value '" . self::TOKEN_VALUE . "' can also be found on " . self::TOKEN_LINK . ".";

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected EntityTypeManagerInterface $entityTypeManager;

  /**
   * Constructs a UniqueFieldAjaxFormFieldsService.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   */
  public function __construct(EntityTypeManagerInterface $entity_type_manager) {
    $this->entityTypeManager = $entity_type_manager;
  }

  /**
   * Return the form field information.
   *
   * @param \Drupal\field\Entity\FieldConfig|\Drupal\node\Entity\NodeType $object
   *   Field details.
   *
   * @return array
   *   Form fields.
   */
  protected function formFields($object): array {
    $inputLookup = self::UNIQUE_THIRD_PARTY_SETTINGS_KEY;

    $tokens = "<br/><strong>" . $this->t('The following tokens that can be used:') . "</strong>";
    $tokens .= "<br/>" . $this->t("<em>%link</em> display a link to the entity matching the unique value.", [
      '%link' => self::TOKEN_LINK,
    ]);
    $tokens .= "<br/>" . $this->t("<em>%label</em> display the field label.", [
      '%label' => self::TOKEN_LABEL,
    ]);
    $tokens .= "<br/>" . $this->t("<em>%value</em> display the field value.", [
      '%value' => self::TOKEN_VALUE,
    ]);

    $fields[self::FIELD_UNIQUE] = [
      '#type' => 'checkbox',
      '#title' => $this->t("Unique"),
      '#description' => $this->t("Enable unique options."),
      '#default_value' => $object->getThirdPartySetting(self::UNIQUE_THIRD_PARTY_SETTINGS_KEY, self::FIELD_UNIQUE),
      '#weight' => -10,
    ];

    $fields[self::FIELD_PER_LANG] = [
      '#type' => 'checkbox',
      '#title' => $this->t("Per Language"),
      '#description' => $this->t("Do not allow duplicated content per language."),
      '#default_value' => $object->getThirdPartySetting(self::UNIQUE_THIRD_PARTY_SETTINGS_KEY, self::FIELD_PER_LANG),
      '#weight' => -9,
      '#states' => [
        'visible' => [
          ':input[name="' . $inputLookup . '[' . self::FIELD_UNIQUE . ']"]' => ['checked' => TRUE],
        ],
      ],
    ];

    $fields[self::FIELD_CASE_SENSITIVE] = [
      '#type' => 'checkbox',
      '#title' => $this->t("Case Sensitive"),
      '#description' => $this->t("Matches will be done on the exact case."),
      '#default_value' => $object->getThirdPartySetting(self::UNIQUE_THIRD_PARTY_SETTINGS_KEY, self::FIELD_CASE_SENSITIVE),
      '#weight' => -9,
      '#states' => [
        'visible' => [
          ':input[name="' . $inputLookup . '[' . self::FIELD_UNIQUE . ']"]' => ['checked' => TRUE],
        ],
      ],
    ];

    $fields[self::FIELD_USE_AJAX] = [
      '#type' => 'checkbox',
      '#title' => $this->t("Use Ajax"),
      '#description' => $this->t("Use ajax for validation."),
      '#default_value' => $object->getThirdPartySetting(self::UNIQUE_THIRD_PARTY_SETTINGS_KEY, self::FIELD_USE_AJAX),
      '#weight' => -8,
      '#states' => [
        'visible' => [
          ':input[name="' . $inputLookup . '[' . self::FIELD_UNIQUE . ']"]' => ['checked' => TRUE],
        ],
      ],
    ];

    $fields[self::FIELD_NO_ENFORCE] = [
      '#type' => 'checkbox',
      '#title' => $this->t("Don't enforce uniqueness"),
      '#description' => $this->t("Will display a warning but still allow saving the field."),
      '#default_value' => $object->getThirdPartySetting(self::UNIQUE_THIRD_PARTY_SETTINGS_KEY, self::FIELD_NO_ENFORCE),
      '#weight' => -8,
      '#states' => [
        'visible' => [
          ':input[name="' . $inputLookup . '[' . self::FIELD_UNIQUE . ']"]' => ['checked' => TRUE],
          ':input[name="' . $inputLookup . '[' . self::FIELD_USE_AJAX . ']"]' => ['checked' => TRUE],
        ],
      ],
    ];

    $description = $this->t("The custom warning message when the value is not unique and dont enforce uniqueness is enabled.");
    $description .= $tokens;
    $fields[self::FIELD_MESSAGE_WARNING] = [
      '#type' => 'textarea',
      '#title' => $this->t("Warning message"),
      '#description' => $description,
      '#default_value' => $object->getThirdPartySetting(self::UNIQUE_THIRD_PARTY_SETTINGS_KEY, self::FIELD_MESSAGE_WARNING) ?? self::DEFAULT_WARNING_MESSAGE,
      '#weight' => -7,
      '#states' => [
        'visible' => [
          ':input[name="' . $inputLookup . '[' . self::FIELD_UNIQUE . ']"]' => ['checked' => TRUE],
          ':input[name="' . $inputLookup . '[' . self::FIELD_NO_ENFORCE . ']"]' => ['checked' => TRUE],
        ],
      ],
    ];

    $description = $this->t("The custom error message when the value is not unique.");
    $description .= $tokens;
    $fields[self::FIELD_MESSAGE] = [
      '#type' => 'textarea',
      '#title' => $this->t("Error message"),
      '#description' => $description,
      '#default_value' => $object->getThirdPartySetting(self::UNIQUE_THIRD_PARTY_SETTINGS_KEY, self::FIELD_MESSAGE) ?? self::DEFAULT_ERROR_MESSAGE,
      '#weight' => -7,
      '#states' => [
        'visible' => [
          ':input[name="' . $inputLookup . '[' . self::FIELD_UNIQUE . ']"]' => ['checked' => TRUE],
        ],
      ],
    ];

    return $fields;
  }

  /**
   * Returns the group tab settings information.
   *
   * @return array
   *   Group tab data.
   */
  protected function getGroupTabSettings($is_entity = FALSE): array {
    return [
      '#type' => 'details',
      '#group' => 'additional_settings',
      '#title' => $is_entity ? $this->t('Unique title ajax settings') : $this->t('Unique field ajax settings'),
      '#open' => FALSE,
      '#tree' => TRUE,
    ];
  }

  /**
   * Helper method to set up the correct format for fields and tabs.
   */
  protected function generateFieldStructure(array $fields, $is_entity = FALSE): array {
    $output[self::UNIQUE_THIRD_PARTY_SETTINGS_KEY] = $this->getGroupTabSettings($is_entity);
    if (!empty($fields)) {
      foreach ($fields as $name => $details) {
        $output[self::UNIQUE_THIRD_PARTY_SETTINGS_KEY][$name] = $details;
      }
    }
    return $output;
  }

  /**
   * Validates third party settings and if any missing fills in the gaps.
   */
  protected function validateThirdPartySettings(array &$thirdPartySettings) {
    $allKeys = $this->getAllFieldKeys();
    foreach ($allKeys as $key) {
      if (!isset($thirdPartySettings[$key])) {
        $thirdPartySettings[$key] = NULL;
      }
    }
    // Not required, just makes it visually easier to compare settings.
    ksort($thirdPartySettings);
  }

  /**
   * Generates and configures the element widget data.
   */
  protected function createAndSetElementWidgetData(array &$element, $field_definition, array $overrides = []): void {
    if (!$field_definition) {
      return;
    }

    $thirdPartySettings = $field_definition->getThirdPartySettings(self::UNIQUE_THIRD_PARTY_SETTINGS_KEY);
    if (!empty($thirdPartySettings)) {
      $extraSettings = [
        self::TPS_FIELD_NAME => $overrides[self::TPS_FIELD_NAME] ?? $field_definition->getFieldStorageDefinition()->getName(),
        self::TPS_FIELD_LABEL => $overrides[self::TPS_FIELD_LABEL] ?? $field_definition->getLabel(),
        self::TPS_MAIN_PROPERTY_NAME => $overrides[self::TPS_MAIN_PROPERTY_NAME] ?? $field_definition->getFieldStorageDefinition()->getMainPropertyName(),
      ];
      // Merge third party settings with extra settings.
      $thirdPartySettings = array_merge($extraSettings, $thirdPartySettings);
      // Validate the settings (will fill in any missing gaps).
      $this->validateThirdPartySettings($thirdPartySettings);

      // If unique is enabled then we can continue.
      if ($thirdPartySettings[self::FIELD_UNIQUE] === 1) {
        // Extract main property name so we can reuse it.
        $main_property_name = $thirdPartySettings[self::TPS_MAIN_PROPERTY_NAME];

        // Add all the third party settings to the element.
        $element[$main_property_name]['#' . self::UNIQUE_THIRD_PARTY_SETTINGS_KEY] = $thirdPartySettings;

        // If no enforce is disabled, then add a validator to prevent save.
        if ($thirdPartySettings[self::FIELD_USE_AJAX] !== 1 || !$thirdPartySettings[self::FIELD_NO_ENFORCE]) {
          $element[$main_property_name]['#element_validate'][] = '_unique_field_ajax_validate_unique';
        }

        // If "use ajax" is enabled, add a process to trigger ajax events.
        if ($thirdPartySettings[self::FIELD_USE_AJAX] === 1) {
          $element['#process'] = ['_unique_field_ajax_enable_ajax_events_process'];
        }
      }
    }
  }

  /**
   * Starts the process of altering an element widget.
   */
  public function elementWidgetFormAlter(array &$element, FormStateInterface $form_state, array $context): bool {
    $field_definition = $context['items']->getFieldDefinition();

    // Checks if is a field, then will attach data if anything found.
    if ($field_definition instanceof FieldConfig) {
      $this->createAndSetElementWidgetData($element, $field_definition);
    }

    // Will try and find settings to enable unique on the title.
    if ($field_definition instanceof BaseFieldDefinition || $field_definition instanceof BaseFieldOverride) {
      $entity = NULL;
      // Support for IEF.
      if (isset($context['form']) && isset($context['form']['#type']) && $context['form']['#type'] == 'inline_entity_form') {
        $entity = $context['form']['#default_value'];
      }
      else {
        $form_object = $form_state->getFormObject();
        if (method_exists($form_object, 'getEntity')) {
          $entity = $form_object->getEntity();
        }
      }
      if (!$entity) {
        return FALSE;
      }

      $entity_type = $entity->getEntityType();
      if (!($entity_type instanceof ContentEntityType) || !$entity_type->hasKey('label')) {
        return FALSE;
      }

      $entity_label = $entity_type->getKey('label');
      $name = $field_definition->getFieldStorageDefinition()->getName();
      if ($entity_label !== $name) {
        return FALSE;
      }

      $bundle_def_id = $entity_type->getBundleEntityType();
      if (!$bundle_def_id) {
        return FALSE;
      }

      $bundle_def = NULL;
      try {
        $bundle_def = $this->entityTypeManager->getStorage($bundle_def_id)->load($entity->bundle());
      }
      catch (InvalidPluginDefinitionException | PluginNotFoundException $e) {
        // @todo log this error.
        // Error getting bundle.
      }
      $this->createAndSetElementWidgetData($element, $bundle_def,
        [
          self::TPS_FIELD_NAME => $name,
          self::TPS_FIELD_LABEL => $entity_label,
          self::TPS_MAIN_PROPERTY_NAME => 'value',
        ]
      );
    }

    return TRUE;
  }

  /**
   * Return the form field information for a field.
   *
   * @param \Drupal\field\Entity\FieldConfig $field
   *   Field details.
   *
   * @return array
   *   Form fields.
   */
  public function getFormFieldsForField(FieldConfig $field): array {
    $fields = $this->formFields($field);
    return ['third_party_settings' => $this->generateFieldStructure($fields)];
  }

  /**
   * Return the form field information for an entity.
   *
   * @param \Drupal\node\Entity\NodeType $entity
   *   Field details.
   *
   * @return array
   *   Form fields.
   */
  public function getFormFieldsForEntity(NodeType $entity): array {
    $fields = $this->formFields($entity);
    return $this->generateFieldStructure($fields, TRUE);
  }

  /**
   * Saves the unique field ajax data onto an entity or field.
   *
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   Form state.
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   Entity object.
   */
  public function saveOrAdjustThirdPartyFieldData(FormStateInterface $form_state, EntityInterface $entity): void {
    $fields = $this->getAllFieldKeys();
    foreach ($fields as $field) {
      $value = $form_state->getValue(['unique_field_ajax', $field]);
      if ($value) {
        $entity->setThirdPartySetting('unique_field_ajax', $field, $value);
      }
      else {
        $entity->unsetThirdPartySetting('unique_field_ajax', $field);
      }
    }
  }

  /**
   * Checks if a form id is valid for a unique title data.
   *
   * @param string $form_id
   *   A form id.
   *
   * @return bool
   *   If form is valid.
   */
  public function isValidEntityTitleForm(string $form_id): bool {
    $entity_config_forms = [
      'node_type_edit_form',
      'node_type_add_form',
    ];
    return in_array($form_id, $entity_config_forms);
  }

  /**
   * Returns an array containing all the valid field keys.
   */
  public function getAllFieldKeys(): array {
    // Uses Reflection to get the constants and then field for only fields.
    $reflector = new \ReflectionClass($this);
    return array_filter($reflector->getConstants(), function ($key) {
      return strpos($key, 'FIELD_') === 0;
    }, ARRAY_FILTER_USE_KEY);
  }

}
