<?php

namespace Drupal\unique_field_ajax;

use Drupal\Core\Cache\CacheableMetadata;
use Drupal\Core\Entity\EntityBase;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Language\LanguageInterface;
use Drupal\Core\Render\Markup;
use Drupal\Core\StringTranslation\StringTranslationTrait;

/**
 * Handles common tasks Unique Field Ajax.
 */
class UniqueFieldAjaxService {
  use StringTranslationTrait;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected EntityTypeManagerInterface $entityTypeManager;

  /**
   * Form Fields Service.
   *
   * @var \Drupal\unique_field_ajax\UniqueFieldAjaxFormFieldsService
   */
  protected UniqueFieldAjaxFormFieldsService $ufaFormFieldsService;

  /**
   * Constructs a UniqueFieldAjaxService.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\unique_field_ajax\UniqueFieldAjaxFormFieldsService $unique_field_ajax_form_fields_service
   *   The entity type manager.
   */
  public function __construct(
    EntityTypeManagerInterface $entity_type_manager,
    UniqueFieldAjaxFormFieldsService $unique_field_ajax_form_fields_service
  ) {
    $this->entityTypeManager = $entity_type_manager;
    $this->ufaFormFieldsService = $unique_field_ajax_form_fields_service;
  }

  /**
   * Validates if a data is unique based on incoming settings.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   */
  public function isUnique(EntityBase $entity, array $element, array $langcode, string $field_value): array {
    if (!isset($element['#value'])) {
      $element['#value'] = $field_value;
    }

    $settings = $element["#" . $this->ufaFormFieldsService::UNIQUE_THIRD_PARTY_SETTINGS_KEY];

    $field_name = $settings[$this->ufaFormFieldsService::TPS_FIELD_NAME];
    $is_case_sensitive = $settings[$this->ufaFormFieldsService::FIELD_CASE_SENSITIVE];
    $is_unique_per_lang = $settings[$this->ufaFormFieldsService::FIELD_PER_LANG];

    $bundle = $entity->bundle();

    $entity_type = $entity->getEntityTypeId();
    $entity_type_definition = $this->entityTypeManager->getDefinition($entity_type);
    $entity_type_storage = $this->entityTypeManager->getStorage($entity_type);

    $query = $entity_type_storage->getQuery()->accessCheck(FALSE)->range(0, 1);

    // Switch the operator depending on case sensitivity.
    $operator = $is_case_sensitive ? 'LIKE BINARY' : '=';
    $query->condition($field_name, $field_value, $operator);

    // Ignore the current id (so we don't test for ourselves).
    if (!$entity->isNew()) {
      $query->condition($entity_type_definition->getKey('id'), $entity->id(), '<>');
    }

    // Test if the entity has a bundle.
    if (!empty($entity_type_definition->getKey('bundle'))) {
      $query->condition($entity_type_definition->getKey('bundle'), $bundle,
        '=');
    }

    // Test unique per language.
    if ($is_unique_per_lang) {
      $lang_code = LanguageInterface::LANGCODE_NOT_SPECIFIED;
      if (isset($langcode[0]) && isset($langcode[0]['value'])) {
        $lang_code = $langcode[0]['value'];
      }
      $query->condition('langcode', $lang_code);
    }

    // All the metadata.
    $metadata = $settings;
    $metadata['entity'] = $entity;
    $metadata['element'] = $element;
    $metadata['field_value'] = $field_value;

    // Add the 'unique_field_ajax' query tag and metadata for altering purposes.
    $query->addTag('unique_field_ajax');
    foreach ($metadata as $key => $value) {
      $query->addMetaData($key, $value);
    }

    $entities = $query->execute();

    // Invoke the unique results alter.
    \Drupal::moduleHandler()->invokeAll('unique_field_ajax_unique_results_alter',
        [&$entities, $metadata]
      );

    $entity_ids = !empty($entities) ? array_values($entities) : "";
    $message = $this->getMessage($entity, $element, $entity_ids);

    return [
      'unique' => empty($entities),
      'entities' => $entity_ids,
      'message' => $message,
    ];
  }

  /**
   * Gets the warning or error message.
   */
  public function getMessage(EntityBase $entity, array $element, $entity_ids): array {
    $settings = $element["#" . $this->ufaFormFieldsService::UNIQUE_THIRD_PARTY_SETTINGS_KEY];
    $field_label = $settings[$this->ufaFormFieldsService::TPS_FIELD_LABEL];
    $no_enforce = $settings[$this->ufaFormFieldsService::FIELD_USE_AJAX] ? $settings[$this->ufaFormFieldsService::FIELD_NO_ENFORCE] : 0;
    $message_warning = $settings[$this->ufaFormFieldsService::FIELD_MESSAGE_WARNING] ?: $this->ufaFormFieldsService::DEFAULT_WARNING_MESSAGE;
    $message_error = $settings[$this->ufaFormFieldsService::FIELD_MESSAGE] ?: $this->ufaFormFieldsService::DEFAULT_ERROR_MESSAGE;
    $message = $no_enforce ? $message_warning : $message_error;
    $class_name = $no_enforce ? 'warning' : 'error';
    $value = $element['#value'] ?? '';

    // If message contains the string %link look up the matching entities.
    if (strpos($message, $this->ufaFormFieldsService::TOKEN_LINK) !== FALSE) {
      $replacement = '<em>' . $this->t('Unknown entity') . '</em>';

      if ($entity_ids) {
        $entity_id = $entity_ids[0];
        $entity_with_value = \Drupal::entityTypeManager()
          ->getStorage($entity->getEntityTypeId())
          ->load($entity_id);

        if ($entity_with_value && $entity_with_value->access('view')) {
          $link = $entity_with_value->toLink(NULL, 'canonical', [
            'attributes' => [
              '_target' => 'blank',
            ],
          ]);
          $replacement = $link->toString();
          $link
            ->getUrl()
            ->toString(TRUE)
            ->applyTo($element);
          CacheableMetadata::createFromObject($entity_with_value)
            ->applyTo($element);
        }
      }

      $message = str_replace($this->ufaFormFieldsService::TOKEN_LINK, $replacement, $message);
    }

    // Replace %label with the field label.
    if (strpos($message, $this->ufaFormFieldsService::TOKEN_LABEL) !== FALSE) {
      $message = str_replace($this->ufaFormFieldsService::TOKEN_LABEL, $field_label, $message);
    }

    // Replace %value with the field value.
    if (strpos($message, $this->ufaFormFieldsService::TOKEN_VALUE) !== FALSE) {
      $message = str_replace($this->ufaFormFieldsService::TOKEN_VALUE, $value, $message);
    }

    return [
      'value' => Markup::create(nl2br($message)),
      'class' => $class_name,
    ];
  }

}
